module Lib
    ( runner, wordCleaner, mapWordsSeq, trieBuilder, completeWords
    ) where

import Data.Char (toLower, isAlpha)
import Data.Map as M
import Data.List as L
import Data.Trie as T
import qualified Data.ByteString.Char8 as C
import Data.Ord as O
import Data.Maybe as Mb
import System.IO(IOMode(WriteMode), hPrint, openFile, hClose)
import System.IO.Error(catchIOError, isUserError, ioeGetFileName, isDoesNotExistError)
import System.Exit(die)

mapWordsSeq :: (Foldable t, Ord k) => t k -> Map k Int
mapWordsSeq = L.foldr (\v s->insertWith (+) v 1 s) M.empty

wordCleaner :: [[Char]] -> [[Char]]
wordCleaner x = L.map (L.map toLower . L.filter isAlpha) x

-- Conventional map reduce
-- mapWords :: [String] -> [(String, Int)]
-- mapWords cw = L.map (\x -> (x,1)) cw
-- mapReduce :: (Ord k) => [(k, Int)] -> Map k Int
-- mapReduce x = (M.fromListWith (+) x)

trieBuilder :: [(String, Int)] -> Trie Int
trieBuilder w = T.fromList (L.map (\x -> (C.pack(fst x), -(snd x))) w)

getK :: Ord b => [(C.ByteString, b)] -> Int -> [Char] -> [Char]
getK t k w = kWord
    where
        kWord = if length l > k then C.unpack (fst (l !! k !! 0)) else w
        l = groupBy (\x y -> snd x == snd y) (L.sortBy (O.comparing snd) t)

wordComplete :: Ord a => [Char] -> Int -> Trie a -> [Char]
wordComplete w k trie = kWord
    where
        resTrie = T.toList (T.submap (C.pack w) trie)
        notPresent = isNothing (T.lookup (C.pack w) trie) 
        kWord = if (length resTrie > 0 && notPresent) then getK resTrie k w else w

completeWords :: Ord a => [[Char]] -> Int -> Trie a -> [[Char]]
completeWords ws k trie = [wordComplete w k trie | w <- ws]

saveFile :: Show a => a -> IO ()
saveFile x = do
    outh <- openFile "output.txt" WriteMode
    hPrint outh x
    hClose outh

runner :: FilePath -> Int -> FilePath -> IO ()
runner inpFile k destFile = do
    content <- readFile inpFile
    let w = words content
    let cw = wordCleaner w

    let wordCount = M.toList (mapWordsSeq cw)
    let trie = trieBuilder wordCount

    dest <- readFile destFile
    let ow = words dest
    let ocw = wordCleaner ow

    let res = completeWords ocw k trie
    saveFile res

  `catchIOError` \ e -> die $ case ioeGetFileName e of
    Just fn | isDoesNotExistError e -> "autocomplete: " ++ fn ++ ": openFile: does not exist (No such file or directory)"
    _       | isUserError e -> "Usage: autocomplete <dict-filename> <k> <broken-filename>"
            | otherwise -> show e
