module Lib
    ( runner, wordCleaner, mapWordsSeq, trieBuilder, completeWords
    ) where

import Data.Char (toLower, isAlpha)
import Data.Map as M
import Data.List as L
import Data.Trie as T
import Control.Parallel.Strategies
import qualified Data.ByteString.Char8 as C
import Data.Ord as O
import Data.List.Split (chunksOf)
import Data.Maybe as Mb
import System.IO(IOMode(WriteMode), hPrint, openFile, hClose)
import System.IO.Error(catchIOError, isUserError, ioeGetFileName, isDoesNotExistError)
import System.Exit(die)

parallelEval :: (a -> b) -> [a] -> Eval [b]
parallelEval _ [] = return []
parallelEval f (x:xs) = 
    do
        y <- rpar (f x)
        ys <- parallelEval f xs
        return (y:ys)

parallelComplete :: (t1 -> t2 -> t3 -> a) -> [t1] -> t2 -> t3 -> Eval [a]
parallelComplete _ [] _ _ = return []
parallelComplete f (x:xs) k trie = 
    do
        y <- rpar (f x k trie)
        ys <- parallelComplete f xs k trie
        return (y:ys)

-- Parallel complete with cache -- Approach 2
{-
parallelCompCache :: (t1 -> t2 -> t3 -> t4 -> a) -> [t1] -> t2 -> t3 -> t4 -> Eval [a]
parallelCompCache _ [] _ _ _ = return []
parallelCompCache f (x:xs) k trie cache = 
    do
        y <- rpar (f x k trie cache)
        ys <- parallelCompCache f xs k trie cache
        return (y:ys)
-}

-- Different order reduce
mapWordsSeq :: (Foldable t, Ord k) => t k -> Map k Int
mapWordsSeq = L.foldr (\v s->insertWith (+) v 1 s) M.empty

mergeMapSeq :: (Foldable t, Ord k) => t (Map k Int) -> [(k, Int)]
mergeMapSeq m = M.toList (L.foldr (M.unionWith (+)) M.empty m)

-- Conventional Map reduce -- Approach 2
{-
mapWords :: [String] -> [(String, Int)]
mapWords cw = L.map (\x -> (x,1)) cw

mapReduce :: (Ord k) => [(k, Int)] -> Map k Int
mapReduce x = (M.fromListWith (+) x)
-}

mergeMaps :: (Foldable t, Ord k) => t (Map k Int) -> Map k Int
mergeMaps = L.foldr (M.unionWith (+)) M.empty

wordCleaner :: [[Char]] -> [[Char]]
wordCleaner x = L.map (L.map toLower . L.filter isAlpha) x

trieBuilder :: [(String, Int)] -> Trie Int
trieBuilder w = T.fromList (L.map (\x -> (C.pack(fst x), -(snd x))) w)

mergeTries :: (Foldable t, Num a) => t (Trie a) -> Trie a
mergeTries t = L.foldr (T.mergeBy (\x y -> Just (x + y))) T.empty t

getK :: Ord b => [(C.ByteString, b)] -> Int -> [Char] -> [Char]
getK t k w = kWord
    where
        kWord = if length l > k then C.unpack (fst (l !! k !! 0)) else w
        l = groupBy (\x y -> snd x == snd y) (L.sortBy (O.comparing snd) t)

wordComplete :: Ord a => [Char] -> Int -> Trie a -> [Char]
wordComplete w k trie = kWord
    where
        resTrie = T.toList (T.submap (C.pack w) trie)
        notPresent = isNothing (T.lookup (C.pack w) trie) 
        kWord = if (length resTrie > 0 && notPresent) then getK resTrie k w else w

completeWords :: Ord a => [[Char]] -> Int -> Trie a -> [[Char]]
completeWords ws k trie = [wordComplete w k trie | w <- ws]

-- Approach 2 with cache
{-
mapRes :: Ord k => k -> Map k [Char] -> [Char]
mapRes w cache = Mb.fromMaybe "" (M.lookup w cache)

compWithCache :: Ord a => [[Char]] -> Int -> Trie a -> Map [Char] [Char] -> [[Char]]
compWithCache [] _ _ _ = []
compWithCache (w:ws) k trie cache
    | ws == [] = [v]
    | otherwise = v : compWithCache ws k trie nMap
    where
        comp = wordComplete w k trie
        v = if res == "" then comp else res
        nMap = if res == "" then M.insert w comp cache else cache
        res = mapRes w cache
-}

saveFile :: Show a => a -> IO ()
saveFile x = do
    outh <- openFile "output.txt" WriteMode
    hPrint outh x
    hClose outh

runner :: FilePath -> Int -> FilePath -> IO ()
runner inpFile k destFile = do
    content <- readFile inpFile
    let w = words content
    let cw = wordCleaner w

    let aChunk = 20000
    let bChunk = 50
    let cChunk = 300
    let dChunk = 50
    let eChunk = 1
    let wordMap   = runEval (parallelEval mapWordsSeq (chunksOf aChunk cw))
    let wordMerge = runEval (parallelEval mergeMapSeq (chunksOf bChunk wordMap))
    let wordCount = M.toList (mergeMaps (L.map M.fromList wordMerge))

    -- Conventional map reduce -- Approach 2
    -- let wordMap = runEval (parallelEval mapWords (chunksOf 10000 cw))
    -- let wordMerge = runEval (parallelEval mapReduce wordMap)
    -- let wordCount = M.toList (mergeMaps wordMerge)

    let wordRes = runEval (parallelEval trieBuilder (chunksOf cChunk wordCount))
    let trieMerge = runEval (parallelEval mergeTries (chunksOf dChunk wordRes))
    let trie = mergeTries trieMerge

    dest <- readFile destFile
    let ow = words dest
    let ocw = wordCleaner ow
    
    -- With a word cache -- Approach 2
    -- let res = concat $ runEval (parallelCompCache compWithCache (chunksOf 5000 ocw) k trie M.empty)

    let res = concat $ runEval (parallelComplete completeWords (chunksOf eChunk ocw) k trie)
    saveFile res

  `catchIOError` \ e -> die $ case ioeGetFileName e of
    Just fn | isDoesNotExistError e -> "autocomplete: " ++ fn ++ ": openFile: does not exist (No such file or directory)"
    _       | isUserError e -> "Usage: autocomplete <dict-filename> <k> <broken-filename>"
            | otherwise -> show e